#include "FreeRTOS.h"
#include "task.h"

/* SDK */
#include "n32wb03x.h"

/* BLE */
#include "ns_ble.h"
#include "ns_ble_task.h"
#include "ns_sleep.h"
#include "rwip.h"
#include "rwip_int.h"

#define MAX_SLEEP_DURATION_PERIODIC_WAKEUP 0x0640
#define MAX_SLEEP_DURATION_EXTERNAL_WAKEUP 0x7D00

extern uint32_t ns_sleep_lock;
extern volatile uint16_t *p_prevent_sleep;

extern void entry_sleep(void);

static uint8_t app_check_stack_status(uint32_t msec);

void vPortSuppressTicksAndSleep(TickType_t xExpectedIdleTime) {
    rwip_time_t sleep_start, sleep_end;

    if (ns_sleep_lock == 1) {
        return;
    }

    /* Stop SysTick timer */
    SysTick->CTRL &= ~(SysTick_CTRL_ENABLE_Msk | SysTick_CTRL_TICKINT_Msk);

    /* Disable global interrupts */
    vPortEnterCritical();

    sleep_start = rwip_time_get();

    /* User supplied routine before low-power state entry */
    app_sleep_prepare_proc();

    /* Check whether BLE stack allows SLEEP */
    if (app_check_stack_status(xExpectedIdleTime) == RWIP_DEEP_SLEEP) {
        entry_sleep();
    }

    vPortExitCritical();

    /* User supplied routine after low-power state entry */
    app_sleep_resume_proc();

    /* Check whether BLE stack wake up completed */
    while ((*p_prevent_sleep & (RW_WAKE_UP_ONGOING | RW_DEEP_SLEEP))) {
        /* -- */
    }

    sleep_end = rwip_time_get();

    if (rwip_env.timer_1ms_target.hs != RWIP_INVALID_TARGET_TIME) {
        if (CLK_DIFF(sleep_end.hs, rwip_env.timer_1ms_target.hs) < 0) {
            ke_event_set(KE_EVENT_KE_TIMER);
        }
    }

    /* Fix-up actual sleep time */
    uint32_t elapsed_ms;
    if (sleep_end.hs == 0) {
        elapsed_ms = xExpectedIdleTime;
        goto step_tick_exit;
    }

    /* Elapsed time in half-slots */
    int32_t elapsed_hs = CLK_DIFF(sleep_start.hs, sleep_end.hs);
    if (elapsed_hs < 0) {
        elapsed_ms = xExpectedIdleTime;
        goto step_tick_exit;
    }

    elapsed_ms = (uint32_t)(elapsed_hs + 1) * HALF_SLOT_SIZE / 2000;
    if (elapsed_ms > xExpectedIdleTime) {
        elapsed_ms = xExpectedIdleTime;
    }

step_tick_exit:
    vTaskStepTick(elapsed_ms);

    /* Re-enable SysTick */
    SysTick->CTRL |= SysTick_CTRL_ENABLE_Msk | SysTick_CTRL_TICKINT_Msk;
}

void vApplicationMallocFailedHook(void) {
    for (;;) {
        __WFI();
    }
}

void vApplicationStackOverflowHook(TaskHandle_t xTask, char *pcTaskName) {
    for (;;) {
        __WFI();
    }
}

/**
 * Supplied by Nations, See LICENSE headers in SDK files.
 * @param msec
 * @return
 */
static uint8_t app_check_stack_status(uint32_t msec) {
    uint8_t sleep_res = RWIP_ACTIVE;

    do {
        if (g_sleep_sleep_enable == 0)  // sleep condition param  0: not sleep 1: allow to sleep
        {
            break;
        }

#if (BLE_EMB_PRESENT || BT_EMB_PRESENT)
        int32_t     sleep_duration;
        rwip_time_t current_time;
#endif  // (BLE_EMB_PRESENT || BT_EMB_PRESENT)

        /************************************************************************
         **************            CHECK KERNEL EVENTS             **************
         ************************************************************************/

        // Check if some kernel processing is ongoing (during wakeup, kernel events are not processed)
        if (((*p_prevent_sleep & RW_WAKE_UP_ONGOING) == 0) && !ke_sleep_check()) {
            break;
        }

        // Processor sleep can be enabled
        sleep_res = RWIP_CPU_SLEEP;

        /************************************************************************
         **************              CHECK RW FLAGS                **************
         ************************************************************************/
        // First check if no pending procedure prevent from going to sleep
        if (*p_prevent_sleep != 0) {
            break;
        }

#if (BLE_EMB_PRESENT || BT_EMB_PRESENT)
        /************************************************************************
         **************           Retrieve Current time            **************
         ************************************************************************/
        current_time = rwip_time_get();
        // Consider 2 half-slots for clock correction (worst case: 1 half-slot before correction, 1 half-slot after
        // correction)
        current_time.hs += 2;

        // Remove 1 more slot because next slot will be started at end of function
        //        if((HALF_SLOT_INV(current_time.hus)) < rwip_env.sleep_algo_dur)
        { current_time.hs += 1; }
        // Be sure that we don't exceed the clock wrapping time
        current_time.hs &= RWIP_MAX_CLOCK_TIME;

        /************************************************************************
         ******* COMPUTE SLEEP TIME ACCORDING TO 1 MS AND HALF SLOT TIMER ******
         ************************************************************************/

        // put sleep duration to maximum value
        sleep_duration =
            (rwip_env.ext_wakeup_enable) ? MAX_SLEEP_DURATION_EXTERNAL_WAKEUP : MAX_SLEEP_DURATION_PERIODIC_WAKEUP;

        // check if 1ms timer is active
        if (rwip_env.timer_1ms_target.hs != RWIP_INVALID_TARGET_TIME) {
            int32_t duration = CLK_DIFF(current_time.hs, rwip_env.timer_1ms_target.hs);
            // update sleep duration to minimum value
            sleep_duration = co_min_s(sleep_duration, duration);
        }

        // check if Half slot timer is active
        if (rwip_env.timer_hs_target != RWIP_INVALID_TARGET_TIME) {
            int32_t duration = CLK_DIFF(current_time.hs, rwip_env.timer_hs_target);
            // update sleep duration to minimum value
            sleep_duration = co_min_s(sleep_duration, duration);
        }

        // check if Half us timer is active
        if (rwip_env.timer_hus_target != RWIP_INVALID_TARGET_TIME) {
            int32_t duration = CLK_DIFF(current_time.hs, rwip_env.timer_hus_target);
            // update sleep duration to minimum value
            sleep_duration = co_min_s(sleep_duration, duration);
        }

        // for rtos sleep time
        uint32_t rtos_slp_slot = msec * 10000 / 3125;
        sleep_duration         = co_min_s(sleep_duration, (int32_t)rtos_slp_slot);

        // A timer ISR is not yet handled or will be raised soon
        // note the sleep duration could be negative, that's why it's useful to check if a minimum requirement is ok
        // at least one half slot.
        if (sleep_duration <= RWIP_MINIMUM_SLEEP_TIME) {
            break;
        }

        if (app_env.lsc_cfg != BLE_LSC_LSE_32768HZ) {
            if ((RCC->OSCFCSR & 0x01)) {
                uint32_t count_value     = 0;
                count_value              = RCC->OSCFCLSICNT;
                g_lsi_1_syscle_cnt_value = (count_value / g_lsi_count_n_syscle) +
                                           (count_value % g_lsi_count_n_syscle) / (g_lsi_count_n_syscle / 2);
            }
        }
        /************************************************************************
         **************           CHECK SLEEP TIME                 **************
         ************************************************************************/

        sleep_duration = (int32_t)(sleep_duration * 10000 / g_lsi_1_syscle_cnt_value) - 1;
        // check if sleep duration is sufficient according to wake-up delay

        if (sleep_duration < rwip_env.lp_cycle_wakeup_delay + 1) {
            break;
        }

        sleep_res = RWIP_DEEP_SLEEP;

        /************************************************************************
         **************          PROGRAM CORE DEEP SLEEP           **************
         ************************************************************************/
        // Program wake-up time
        ip_deepslwkup_set(sleep_duration);

        // Prevent re-entering sleep, until it effectively sleeps and wakes up
        rwip_prevent_sleep_set(RW_DEEP_SLEEP);

        /************************************************************************
         **************               SWITCH OFF RF                **************
         ************************************************************************/
        rwip_rf.sleep();
#endif  // (BLE_EMB_PRESENT || BT_EMB_PRESENT)
    } while (0);

    return sleep_res;
}
